#include <math.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "mpmath.h"

//---------------------------------------------------------------------
//  mainline
//---------------------------------------------------------------------

int main(int argc, char *argv[])
{
	FILE *fp1;
    short *a,*a2,*b,*b2,*c2,*s,*x,*y;
    char str[80];
    int d=0,i,j,k,n,p,iterations,start,finish;
	double fv;

    if (argc>1)
        sscanf(argv[1],"%ld",&d);

    while (d==0) {
        printf ("enter the number of digits => ");
        scanf ("%s",str );
        sscanf (str,"%ld",&d );
    }

    for (n=8;n<d;n <<= 1); d=n; n >>= 2;
    printf("integers=%ld  digits=%ld\n",n,d);

	fv=(1.0-sqrt(0.5))*d;
	k=(int)(fv+0.5);
	iterations=Log2(k)+1;
	printf("number of iterations = %d\n",iterations);

	//------------
	//  allocate
	//------------

	start=clock();

	printf("Pi2     allocating %9d bytes\n",8*n*sizeof(short));
    a=ivector(n);
    a2=ivector(n);
    b=ivector(n);
    b2=ivector(n);
    c2=ivector(n);
    s=ivector(n);
    x=ivector(n);
    y=ivector(n);

	fullmul(NULL,NULL,NULL,n,0,GRAB);
	mpsqrt(NULL,NULL,NULL,n,GRAB);
	mpinv(NULL,NULL,NULL,n,GRAB);

	finish=clock();
	printf("allocation time   = %10.2f sec\n",(float)(finish-start)/CLOCKS_PER_SEC);

	//-------------
	//  calculate
	//-------------

	start=finish;

	// s[0] = 1
		
	mpclr(s,n);
	s[0]=1;

	// a[0] = 1
	
	mpclr(a,n);
	a[0]=1;

	// a[0]^2 = 1
	
	mpclr(a2,n);
	a2[0]=1;

	// b[0]^2 = 0.5
	
	mpclr(b2,n);
	b2[1]=5000;
	
	// b[0] = sqrt(b[0])
	
	mpsqrt(y,x,b2,n,NOP);
	mpmov(b,y,n);

	for (k=1;k<=iterations;k++) {
	
		printf("Pass k = %d\n",k);
	
		if (k==1) {

			// a[1] = (a[0] + b[0]) / 2

			mpadd(a,a,b,n);
			mpsdv(a,a,n,2);
			
			// b[1]^2 = a[0] * b[0] = b[0]

			mpmov(c2,b2,n);
			mpmov(b2,b,n);
			
			// c[1]^2 = ((a[0]^2 + b[0]^2) / 2 - b[1]^2) / 2

			mpadd(c2,a2,c2,n);
			mpsdv(c2,c2,n,2);
			mpsub(c2,c2,b,n);
			mpsdv(c2,c2,n,2);

		} else {

			// c[n] = (a[n-1] - b[n-1]) / 2
	
			mpsub(c2,a,b,n);
			mpsdv(c2,c2,n,2);

			// a[n] = a[n-1] - c[n]
	
			mpsub(a,a,c2,n);

			// c[n]^2 = c[n] * c[n]
	
			fullmul(c2,c2,c2,n,n,LSH);

			// b[n]^2 = (a[n-1]^2 + b[n-1]^2 - 4 * c[n]^2) / 2
	
			mpsmu(b,c2,n,4);
			mpadd(b2,a2,b2,n);
			mpsub(b2,b2,b,n);
			mpsdv(b2,b2,n,2);
		}

		// a[n]^2 = c[n]^2 + b[n]^2

		mpadd(a2,c2,b2,n);

		// b[n] = sqrt(b[n])
	
		if (k<iterations) {
			mpsqrt(y,x,b2,n,NOP);
			mpmov(b,y,n);
		}

		// s = s - c[n]^2 * 2^(k+1)

		if ((p=k+1)<=13) {
			mpsmu(c2,c2,n,1<<p);
		} else {
			p=p-13;
			mpsmu(c2,c2,n,1<<13);
			mpsmu(c2,c2,n,1<<p);
		}

		mpsub(s,s,c2,n);
	}
	
	printf("Final Pass\n");

	// a[n]^2 = (a[n-1]^2 + b[n-1]^2) / 2

	mpadd(a2,a2,b2,n);
	mpsdv(a2,a2,n,2);

	// Pi = 4 * a[n]^2 / s

	mpsmu(a2,a2,n,4);

	mpinv(y,x,s,n,NOP);
	mpmov(s,y,n);

	fullmul(a,a2,s,n,n,LSH);

	finish=clock();
	printf("calculation time  = %10.2f sec\n",(float)(finish-start)/CLOCKS_PER_SEC);

	//--------------------
	//  write the result
	//--------------------

	start=finish;

    if ((fp1=fopen("pi.txt","w"))==NULL) error("fopen");
	// fprintf(fp1,"%d.\n",a[0]);
	for (i=1,k=0;i<n;i++) {
		sprintf(str,"%04d",a[i]);
		for (j=0;j<4;j++) {
			k++;
			if (k%1000==0)
				fprintf(fp1,"%c\n\n",str[j]);
			else if (k%50==0)
				fprintf(fp1,"%c\n",str[j]);
			else if (k%5==0)
				fprintf(fp1,"%c ",str[j]);
			else
				fprintf(fp1,"%c",str[j]);
		}
	}
	fprintf(fp1,"\n");
	fclose(fp1);

	finish=clock();
	printf("file writing time = %10.2f sec\n",(float)(finish-start)/CLOCKS_PER_SEC);

	//--------------
	//  deallocate
	//--------------

	start=finish;

	mpinv(NULL,NULL,NULL,n,FREE);
	mpsqrt(NULL,NULL,NULL,n,FREE);
	fullmul(NULL,NULL,NULL,n,0,FREE);

    free_ivector(y);
    free_ivector(x);
    free_ivector(s);
    free_ivector(c2);
    free_ivector(b2);
    free_ivector(b);
    free_ivector(a2);
    free_ivector(a);

	finish=clock();
	printf("deallocation time = %10.2f sec\n",(float)(finish-start)/CLOCKS_PER_SEC);
	
    return 0;
}

void show(FILE *fp,char *s,short *u,int n)
{
    int i;

    printf("%s\n",s);
    fprintf(fp,"%s\n",s);
    for (i=0;i<n;i++) {
        if ((i+1)%16==0 && (i+1)!=n) {
            printf("%04d\n",u[i]);
            fprintf(fp,"%04d\n",u[i]);
        } else {
            printf("%04d ",u[i]);
            fprintf(fp,"%04d ",u[i]);
        }
    }
    printf("\n");
    fprintf(fp,"\n");
}
