#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "mpmath.h"

#define BASE 10000

//------------------------------------------------------------------------------
//  mpadd  -  add
//------------------------------------------------------------------------------

short mpadd(short w[], short u[], short v[], int n)
{
    int j,s,c=0;

    for (j=n-1;j>=0;j--) {
        s=u[j]+v[j]+c;
        if (s>=BASE) {
            w[j]=(short) (s-BASE);
            c=1;
        } else {
            w[j]=(short) s;
            c=0;
        }
    }
    return c;
}

//------------------------------------------------------------------------------
//  mpsub  -  subtract
//------------------------------------------------------------------------------

short mpsub(short w[], short u[], short v[], int n)
{
    int j,s,c=0;

    for (j=n-1;j>=0;j--) {
        s=u[j]-v[j]-c;
        if (s<0) {
            w[j]=(short) (s+BASE);
            c=1;
        } else {
            w[j]=(short) s;
            c=0;
        }
    }
    return c;
}

//------------------------------------------------------------------------------
//  mpsad  -  short add
//------------------------------------------------------------------------------

short mpsad(short w[], short u[], int n, int iv)
{
    int j,s,c=iv;

    for (j=n-1;j>=0;j--) {
        s=u[j]+c;
        if (s>=BASE) {
            w[j]=(short) (s-BASE);
            c=1;
        } else {
            w[j]=(short) s;
            c=0;
			break;
        }
    }
    return c;
}

//------------------------------------------------------------------------------
//  mpssu  -  short subtract
//------------------------------------------------------------------------------

short mpssu(short w[], short u[], int n, int iv)
{
    int j,s,c=iv;

    for (j=n-1;j>=0;j--) {
        s=u[j]-c;
        if (s<0) {
            w[j]=(short) (s+BASE);
            c=1;
        } else {
            w[j]=(short) s;
            c=0;
			break;
        }
    }
    return c;
}

//------------------------------------------------------------------------------
//  mpsmu  -  short multiply
//------------------------------------------------------------------------------

short mpsmu(short w[], short u[], int n, int iv)
{
    int j,s,c=0;

    for (j=n-1;j>=0;j--) {
        s=u[j]*iv+c;
        w[j]=(short) (s % BASE);
        c=s/BASE;
    }
    return c;
}

//------------------------------------------------------------------------------
//  mpsdv  -  short divide
//------------------------------------------------------------------------------

short mpsdv(short w[], short u[], int n, int iv)
{
    int j,s,r=0;

    for (j=0;j<n;j++) {
        s=BASE*r+u[j];
        w[j]=(short) (s/iv);
        r=s % iv;
    }
    return r;
}

//------------------------------------------------------------------------------
//  mpneg  -  negate
//------------------------------------------------------------------------------

short mpneg(short u[], int n)
{
    int j,s,c=0;

    for (j=n-1;j>=0;j--) {
        s=-u[j]-c;
        if (s<0) {
            u[j]=(short) (s+BASE);
            c=1;
        } else {
            u[j]=(short) s;
            c=0;
        }
    }
    return c ;
}

//------------------------------------------------------------------------------
//  mpmov  -  move  
//------------------------------------------------------------------------------

void mpmov(short u[], short v[], int n)
{
    int j;

    for (j=0;j<n;j++) {
		u[j]=v[j];
	}
}

//------------------------------------------------------------------------------
//  mplsh  -  left shift
//------------------------------------------------------------------------------

void mplsh(short u[], int n)
{
    int j;

    for (j=0;j<n-1;j++) {
        u[j]=u[j+1];
	}
    u[n-1]=0;
}

//------------------------------------------------------------------------------
//  mpclr  -  clear
//------------------------------------------------------------------------------

void mpclr(short u[], int n)
{
    int j;

    for (j=0;j<n;j++) {
        u[j]=0;
	}
}

//------------------------------------------------------------------------------

#define IMAX(x,y) (((x) > (y)) ? (x) : (y))
#define RX 10000.0
#define SWAP(a,b) tempr=(a);(a)=(b);(b)=tempr

void halfmul(short w[], short u[], short v[], int n, int m, int flags)
{
	int h,i,j,k;

	h=n>>1;
    for (i=0;i<h && u[i]==0;i++);
    for (j=0;j<h && v[j]==0;j++);
	k=i+j-1;
	if (k<0) {
		fullmul(w,u,v,h,m,flags|LSH);
	} else {
		fullmul(w+k,u+i,v+j,h,m-k,flags&~LSH);
	    for (i=0;i<k;i++) w[i]=0;
	}
}

void fullmul(short w[], short u[], short v[], int n, int m, int flags)
{
    void drealft(double data[], unsigned long n, int isign);
    static double *a,*b,*c;
    long j,nn;
    double cy,t;

    nn=n<<1;

	if (flags==GRAB) {
		printf("fullmul allocating %9d bytes\n",3*nn*sizeof(double));
	    a=dvector(nn);
	    b=dvector(nn);
	    c=dvector(nn);
		drealft(NULL,nn,GRAB);
		return;
	}
	if (flags==FREE) {
		drealft(NULL,nn,FREE);
	    free_dvector(c);
	    free_dvector(b);
		free_dvector(a);
		return;
	}

	if ((flags & RCL) && (u==v)) {
        for (j=0;j<nn;j++)
            a[j]=c[j];
	} else {
		for (j=0;j<n;j++)
			a[j]=(double)u[j];
		for (j=n;j<nn;j++)
			a[j]=0.0;
		drealft(a,nn,1);
	}

	if (flags & RCL) {
        for (j=0;j<nn;j++)
            b[j]=c[j];
	} else if (u==v) {
        for (j=0;j<nn;j++)
            b[j]=a[j];
    } else {
        for (j=0;j<n;j++)
            b[j]=(double)v[j];
        for (j=n;j<nn;j++)
            b[j]=0.0;
        drealft(b,nn,1);
    }

	if (flags & STO) {
        for (j=0;j<nn;j++)
            c[j]=b[j];
	}

    b[0] *= a[0];
    b[1] *= a[1];
    for (j=2;j<nn;j+=2) {
        b[j]=(t=b[j])*a[j]-b[j+1]*a[j+1];
        b[j+1]=t*a[j+1]+b[j+1]*a[j];
    }
    drealft(b,nn,-1);
    
	cy=0.0;
    for (j=nn-1;j>=0;j--) {
        t=b[j]/(nn>>1)+cy+0.5;
        cy=(unsigned long) (t/RX);
        b[j]=t-cy*RX;
    }
    if (cy >= RX)
        error("cannot happen in mpmul");
	
	if (flags & LSH) {
		for (j=0;j<m;j++)
			w[j]=(short) b[j];
	} else {
		w[0]=(short) cy;
		for (j=1;j<m;j++)
			w[j]=(short) b[j-1];
	}
}

//------------------------------------------------------------------------------

void drealft(double data[], unsigned long n, int isign)
{
    void dfour1(double data[], unsigned long nn, int isign);
    unsigned long i,i1,i2,i3,i4,np1;
    double c1=0.5,c2,h1r,h1i,h2r,h2i;
    double wr,wi,wpr,wpi,wtemp,theta;

	if (isign==GRAB) {
        dfour1(NULL,n>>1,GRAB);
		return;
	}
	if (isign==FREE) {
        dfour1(NULL,n>>1,FREE);
		return;
	}

    theta=3.141592653589793/(double) (n>>1);
    if (isign == 1) {
        c2 = -0.5;
        dfour1(data,n>>1,1);
    } else {
        c2=0.5;
        theta = -theta;
    }
    wtemp=sin(0.5*theta);
    wpr = -2.0*wtemp*wtemp;
    wpi=sin(theta);
    wr=1.0+wpr;
    wi=wpi;
    np1=n+1;
    for (i=2;i<=(n>>2);i++) {
        i4=1+(i3=np1-(i2=1+(i1=i+i-2)));
        h1r=c1*(data[i1]+data[i3]);
        h1i=c1*(data[i2]-data[i4]);
        h2r = -c2*(data[i2]+data[i4]);
        h2i=c2*(data[i1]-data[i3]);
        data[i1]=h1r+wr*h2r-wi*h2i;
        data[i2]=h1i+wr*h2i+wi*h2r;
        data[i3]=h1r-wr*h2r+wi*h2i;
        data[i4] = -h1i+wr*h2i+wi*h2r;
        wr=(wtemp=wr)*wpr-wi*wpi+wr;
        wi=wi*wpr+wtemp*wpi+wi;
    }
    if (isign == 1) {
        data[0] = (h1r=data[0])+data[1];
        data[1] = h1r-data[1];
    } else {
        data[0]=c1*((h1r=data[0])+data[1]);
        data[1]=c1*(h1r-data[1]);
        dfour1(data,n>>1,-1);
    }
}

//------------------------------------------------------------------------------

void dfour1(double data[], unsigned long nn, int isign)
{
#if 1
	#include "fftw.h"
	
	static fftw_plan forwardPlan[21],backwardPlan[21];
	static double *d2;
	unsigned long n,k;
	fftw_complex *in,*out;
	int planIndex;

	planIndex=Log2(nn);
	n=nn*2;

	if (isign==GRAB) {
		printf("dfour1  allocating %9d bytes\n",n*sizeof(double));
		d2=dvector(n);
		while (nn>0) {
			forwardPlan[planIndex]=fftw_create_plan(nn,FFTW_FORWARD,FFTW_ESTIMATE);
			backwardPlan[planIndex]=fftw_create_plan(nn,FFTW_BACKWARD,FFTW_ESTIMATE);
			nn>>=1;
			planIndex--;
		}
		return;
	}
	if (isign==FREE) {
		free_dvector(d2);
		while (nn>0) {
			fftw_destroy_plan(forwardPlan[planIndex]);
			fftw_destroy_plan(backwardPlan[planIndex]);
			nn>>=1;
			planIndex--;
		}
		return;
	}

	in =(fftw_complex*) data;
	out=(fftw_complex*) d2;
	if (isign==1)
		fftw_one(forwardPlan[planIndex],in,out);
	else
		fftw_one(backwardPlan[planIndex],in,out);

	data[0]=out[0].re;
	data[1]=out[0].im;
	for (k=1;k<nn;k++) {
		data[2*k]=out[nn-k].re;
		data[2*k+1]=out[nn-k].im;
	}

#else
    unsigned long n,mmax,m,j,istep,i;
    double wtemp,wr,wpr,wpi,wi,theta;
    double tempr,tempi;

    n=nn << 1;
    j=1;
    for (i=1;i<n;i+=2) {
        if (j > i) {
            SWAP(data[j-1],data[i-1]);
            SWAP(data[j],data[i]);
        }
        m=n >> 1;
        while (m >= 2 && j > m) {
            j -= m;
            m >>= 1;
        }
        j += m;
    }
    mmax=2;
    while (n > mmax) {
        istep=mmax*2;
        theta=isign*(6.28318530717959/mmax);
        wtemp=sin(0.5*theta);
        wpr = -2.0*wtemp*wtemp;
        wpi=sin(theta);
        wr=1.0;
        wi=0.0;
        for (m=1;m<mmax;m+=2) {
            for (i=m;i<=n;i+=istep) {
                j=i+mmax;
                tempr=wr*data[j-1]-wi*data[j];
                tempi=wr*data[j]+wi*data[j-1];
                data[j-1]=data[i-1]-tempr;
                data[j]=data[i]-tempi;
                data[i-1] += tempr;
                data[i] += tempi;
            }
            wr=(wtemp=wr)*wpr-wi*wpi+wr;
            wi=wi*wpr+wtemp*wpi+wi;
        }
        mmax=istep;
    }
#endif
}

//------------------------------------------------------------------------------

short *ivector(long n)
{
    short *v;

    v=(short *)malloc((size_t) (n*sizeof(short)));
    if (!v) error("allocation failure in ivector()");
    return v;
}

//------------------------------------------------------------------------------

void free_ivector(short *v)
{
    free((char*) v);
}

//------------------------------------------------------------------------------

double *dvector(long n)
{
    double *v;

    v=(double *)malloc((size_t) (n*sizeof(double)));
    if (!v) error("allocation failure in dvector()");
    return v;
}

//------------------------------------------------------------------------------

void free_dvector(double *v)
{
    free((char*) v);
}

//------------------------------------------------------------------------------

void error(char error_text[])
{
    fprintf(stderr,"run-time error...\n");
    fprintf(stderr,"%s\n",error_text);
    exit(1);
}

//------------------------------------------------------------------------------

int Log2(int Num)
{
    int x=-1;
    if (Num==0)
        return 0;
    while (Num)
    {
        x++;
        Num/=2;
    }
    return x;
}
