#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpmath.h"

#define IMIN(x,y) (((x) < (y)) ? (x) : (y))
#define MF 3
#define RX 10000.0
#define BI (1.0/RX)


void mpsqrt(short y[], short x[], short a[], int n9, int flags)
{
	static short *u;
    int i,j,mm,isign,n=n9,k,k9;
    double fu,fv;

    if (flags==GRAB) {
		printf("mpsqrt  allocating %9d bytes\n",n*sizeof(short));
	    u=ivector(n);
		return;
	}
    if (flags==FREE) {
		free_ivector(u);
		return;
	}

    mm=IMIN(n,MF);
    fv=(double) a[mm];
//  for (j=mm-1;j>=0;j--) {
    for (j=n9-1;j>=0;j--) {
        fv *= BI;
        fv += a[j];
    }
    fu=1.0/sqrt(fv);
    for (j=0;j<n;j++) {
        i=(int) fu;
        x[j]=(short) i;
        fu=RX*(fu-i);
    }

	for (n=4;n<n9;n*=2) {
		if (n==n9/4) k9=2; else k9=1;
		for (k=0;k<k9;k++) {
    
			//    x = x + x . ( 1 - A . x ^ 2 ) / 2

			//------------------------------------
			//  u = x ^ 2
			//------------------------------------

			fullmul(u,x,x,n,n,LSH+STO);

			//------------------------------------
			//  u = a  u
			//------------------------------------

			fullmul(u,a,u,n,n,LSH);

			//------------------------------------
			//  u = 1 - u
			//------------------------------------

			if (u[0]==0) {
				mpneg(u,n);
				u[0]-=9999;
				isign = +1;
			} else {
				u[0]--;
				isign = -1;
			}

			//------------------------------------
			//  u = x  u
			//------------------------------------

			fullmul(u,u,x,n,n,LSH+RCL);

			//------------------------------------
			//  u = u / 2
			//------------------------------------

			mpsdv(u,u,n,2);

			//------------------------------------
			//  x = x + u
			//------------------------------------

			if (isign == +1)
				mpadd(x,x,u,n);
			else    
				mpsub(x,x,u,n);
		}
	}

    n=n9;

    //  y = A  x
    
    //  y = y + x  ( A - y ^ 2 ) / 2

    //------------------------------------
    //  y = a  x
    //------------------------------------

    halfmul(y,a,x,n,n,STO);
    mpclr(y+n/2,n/2);

    //------------------------------------
    //  u = y  y
    //------------------------------------

    halfmul(u,y,y,n,n,0);
        
    //------------------------------------
    //  u = a - u
    //------------------------------------

    i = mpsub(u,a,u,n);
    if (i==0) {
        isign = +1;
    } else {
        isign = -1;
        mpneg(u,n);
    }

    //------------------------------------
    //  u = x  u
    //------------------------------------

    halfmul(u,u,x,n,n,RCL);

    //------------------------------------
    //  u = u / 2
    //------------------------------------

    mpsdv(u,u,n,2);

    //------------------------------------
    //  y = y + u
    //------------------------------------

    if (isign == +1)
        mpadd(y,y,u,n);
    else
        mpsub(y,y,u,n);

    //------------------------------------
    // end
    //------------------------------------

    return;
}
